#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <time.h>
#include <unistd.h>

typedef struct {
	int value;
	int set_flag;
	int done_flag;
} cell_t ;

static int nb_proc;
static int shmid;

cell_t *vector;

void *create_shared_array(char **argv) {
	key_t key;
	void *ptr;

	key = ftok(argv[0], 'a');
	if (key == -1) {
		fprintf(stderr, "error ftok \n");
		fflush(stderr);
		exit(-1);
	}

	shmid = shmget(key, nb_proc * sizeof(cell_t), IPC_CREAT | 0666);
	if (shmid == -1) {
		perror("shmget");
		fprintf(stderr, "error shmget \n");
		fflush(stderr);
		exit(-1);
	}

	ptr = shmat(shmid, NULL, 0);

	if (ptr == (cell_t *) -1) {
		perror("shmat");
		fprintf(stderr, "error shmat \n", argv[0]);
		fflush(stderr);
		exit(-1);
	}
	return ptr;

}

void delete_shared_memory(void *ptr) {
	int cr;

	cr = shmdt(ptr);
//	fprintf(stderr, "shmdt cr = %d\n", cr);
//	fflush(stderr);

	cr = shmctl(shmid, IPC_RMID, NULL);
//	fprintf(stderr, "shmctl cr = %d\n", cr);
//	fflush(stderr);
}

void node(int me) {
	int my_pid, my_val;
	int i;
	int best_val, best_pid, best_node;

	my_pid = getpid();
	srand(my_pid);
	my_val = rand() % 1024;


	vector[me].value    = my_val;
	vector[me].set_flag = 1;

	printf("Process node %d val = %d\n", me, my_val);

	best_val = my_val;
	best_pid = my_pid;
	best_node = me;

	for (i = 0; i < nb_proc; i++) {
		// wait actively
		while (vector[i].set_flag == 0)
			;

		if (vector[i].value > best_val) {
			best_val = vector[i].value;
			best_node = i;
		}
	}

	fprintf(stderr, "Node %d: the winner is %d node %d\n", me, best_val, best_node);

	vector[me].done_flag = 1;

	if (me == 0) {
		for (i = 0; i < nb_proc; i++) {
			/* wait actively again */
			while (vector[i].done_flag == 0)
				;
		}
		delete_shared_memory(vector);
	}

	exit(0);
}

void init_shared_array() {
	int i;

	for (i = 0; i < nb_proc; i++) {
		vector[i].value     = -1;
		vector[i].set_flag  = 0;
		vector[i].done_flag = 0;
	}
}

int main(int argc, char **argv) {
	int pid;
	int i;

	if (argc != 2) {
		fprintf(stderr, "Usage: ring4 nb\n");
		exit(-1);
	}

	nb_proc = atoi(argv[1]);

	vector = create_shared_array(argv);

	init_shared_array();

	for (i = 1; i < nb_proc; i++) {
		pid = fork();
		if (pid == 0) {
			node(i);
		}
	}
	node(0);

	return 0;
}
